@@ -1,4 +1,4 @@
-/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2021 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -31,11 +31,11 @@
 #include "knot/updates/zone-update.h"
 #include "knot/zone/backup.h"
 #include "knot/zone/timers.h"
+#include "knot/zone/zonedb-load.h"
 #include "knot/zone/zonefile.h"
 #include "libknot/libknot.h"
 #include "libknot/yparser/yptrafo.h"
 #include "contrib/files.h"
-#include "contrib/macros.h"
 #include "contrib/string.h"
 #include "contrib/strtonum.h"
 #include "contrib/ucw/lists.h"
@@ -47,6 +47,13 @@
 #define MATCH_AND_FILTER(args, code) ((args)->data[KNOT_CTL_IDX_FILTER] != NULL && \
                                       strchr((args)->data[KNOT_CTL_IDX_FILTER], (code)) != NULL)
 
+#define RETURN_IF_FAILED(exception) \
+{ \
+	if (ret != KNOT_EOK && ret != (exception)) { \
+		return ret; \
+	} \
+}
+
 typedef struct {
 	ctl_args_t *args;
 	int type_filter; // -1: no specific type, [0, 2^16]: specific type.
@@ -68,16 +75,42 @@ static struct {
 	            sizeof(((send_ctx_t *)0)->rdata)];
 } ctl_globals;
 
-static void schedule_trigger(zone_t *zone, ctl_args_t *args, zone_event_type_t event,
-                             bool user)
+/*!
+ * Evaluates a filter pair and checks for conflicting filters.
+ *
+ * \param[in]  args        Command arguments.
+ * \param[out] param       The filter to be set.
+ * \param[in]  dflt        Default filter value.
+ * \param[in]  filter      Name of the filter.
+ * \param[in]  neg_filter  Name of the negative filter.
+ *
+ * \return false if there is a filter conflict, true otherwise.
+ */
+
+static bool eval_opposite_filters(ctl_args_t *args, bool *param, bool dflt,
+                                  int filter, int neg_filter)
 {
+	bool set = MATCH_AND_FILTER(args, filter);
+	bool unset = MATCH_AND_FILTER(args, neg_filter);
+
+	*param = dflt ? (set || !unset) : (set && !unset);
+	return !(set && unset);
+}
+
+static int schedule_trigger(zone_t *zone, ctl_args_t *args, zone_event_type_t event,
+                            bool user)
+{
+	int ret = KNOT_EOK;
+
 	if (ctl_has_flag(args->data[KNOT_CTL_IDX_FLAGS], CTL_FLAG_BLOCKING)) {
-		zone_events_schedule_blocking(zone, event, user);
+		ret = zone_events_schedule_blocking(zone, event, user);
 	} else if (user) {
 		zone_events_schedule_user(zone, event);
 	} else {
 		zone_events_schedule_now(zone, event);
 	}
+
+	return ret;
 }
 
 void ctl_log_data(knot_ctl_data_t *data)
@@ -148,18 +181,28 @@ static int get_zone(ctl_args_t *args, zone_t **zone)
 
 static int zones_apply(ctl_args_t *args, int (*fcn)(zone_t *, ctl_args_t *))
 {
-	int ret = KNOT_EOK;
+	int ret;
 
 	// Process all configured zones if none is specified.
 	if (args->data[KNOT_CTL_IDX_ZONE] == NULL) {
-		args->failed = false;
-		knot_zonedb_foreach(args->server->zone_db, fcn, args);
-		if (args->failed) {
+		bool failed = false;
+		knot_zonedb_iter_t *it = knot_zonedb_iter_begin(args->server->zone_db);
+		while (!knot_zonedb_iter_finished(it)) {
+			args->suppress = false;
+			ret = fcn((zone_t *)knot_zonedb_iter_val(it), args);
+			if (ret != KNOT_EOK && !args->suppress) {
+				failed = true;
+			}
+			knot_zonedb_iter_next(it);
+		}
+		knot_zonedb_iter_free(it);
+
+		if (failed) {
 			ret = KNOT_CTL_EZONE;
 			log_ctl_error("control, error (%s)", knot_strerror(ret));
 			send_error(args, knot_strerror(ret));
-			args->failed = false;
 		}
+
 		return KNOT_EOK;
 	}
 
@@ -316,53 +359,44 @@ static int zone_status(zone_t *zone, ctl_args_t *args)
 	return KNOT_EOK;
 }
 
-static int zone_reload(zone_t *zone, ctl_args_t *args)
+static int zone_reload(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	if (zone_expired(zone)) {
+		args->suppress = true;
 		return KNOT_ENOTSUP;
 	}
 
-	schedule_trigger(zone, args, ZONE_EVENT_LOAD, true);
+	if (ctl_has_flag(args->data[KNOT_CTL_IDX_FLAGS], CTL_FLAG_FORCE)) {
+		return zone_reload_modules(conf(), args->server, zone->name);
+	}
 
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_LOAD, true);
 }
 
-static int zone_refresh(zone_t *zone, ctl_args_t *args)
+static int zone_refresh(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	if (!zone_is_slave(conf(), zone)) {
+		args->suppress = true;
 		return KNOT_ENOTSUP;
 	}
 
-	schedule_trigger(zone, args, ZONE_EVENT_REFRESH, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_REFRESH, true);
 }
 
-static int zone_retransfer(zone_t *zone, ctl_args_t *args)
+static int zone_retransfer(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	if (!zone_is_slave(conf(), zone)) {
+		args->suppress = true;
 		return KNOT_ENOTSUP;
 	}
 
 	zone_set_flag(zone, ZONE_FORCE_AXFR);
-	schedule_trigger(zone, args, ZONE_EVENT_REFRESH, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_REFRESH, true);
 }
 
-static int zone_notify(zone_t *zone, ctl_args_t *args)
+static int zone_notify(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
-	schedule_trigger(zone, args, ZONE_EVENT_NOTIFY, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_NOTIFY, true);
 }
 
 static int zone_flush(zone_t *zone, ctl_args_t *args)
@@ -374,7 +408,6 @@ static int zone_flush(zone_t *zone, ctl_args_t *args)
 		if (ret != KNOT_EOK) {
 			log_zone_warning(zone->name, "failed to update zone file (%s)",
 			                 knot_strerror(ret));
-			args->failed = true;
 		}
 		return ret;
 	}
@@ -383,19 +416,17 @@ static int zone_flush(zone_t *zone, ctl_args_t *args)
 		zone_set_flag(zone, ZONE_FORCE_FLUSH);
 	}
 
-	schedule_trigger(zone, args, ZONE_EVENT_FLUSH, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_FLUSH, true);
 }
 
 static int init_backup(ctl_args_t *args, bool restore_mode)
 {
-	if (!MATCH_AND_FILTER(args, CTL_FILTER_FLUSH_OUTDIR)) {
+	if (!MATCH_AND_FILTER(args, CTL_FILTER_BACKUP_OUTDIR)) {
 		return KNOT_ENOPARAM;
 	}
 
 	// Make sure that the backup outdir is not the same as the server DB storage.
-	conf_val_t db_storage_val = conf_get(conf(), C_DB, C_STORAGE);
+	conf_val_t db_storage_val = conf_db_param(conf(), C_STORAGE);
 	const char *db_storage = conf_str(&db_storage_val);
 
 	const char *backup_dir = args->data[KNOT_CTL_IDX_DATA];
@@ -413,13 +444,31 @@ static int init_backup(ctl_args_t *args, bool restore_mode)
 		return KNOT_EINVAL;
 	}
 
+	// Evaluate filters (and possibly fail) before writing to the filesystem.
+	bool filter_zonefile, filter_journal, filter_timers, filter_kaspdb, filter_catalog;
+
+	// The default filter values are set just in this paragraph.
+	if (!(eval_opposite_filters(args, &filter_zonefile, true,
+	                            CTL_FILTER_BACKUP_ZONEFILE, CTL_FILTER_BACKUP_NOZONEFILE) &&
+	    eval_opposite_filters(args, &filter_journal, false,
+	                          CTL_FILTER_BACKUP_JOURNAL, CTL_FILTER_BACKUP_NOJOURNAL) &&
+	    eval_opposite_filters(args, &filter_timers, true,
+	                          CTL_FILTER_BACKUP_TIMERS, CTL_FILTER_BACKUP_NOTIMERS) &&
+	    eval_opposite_filters(args, &filter_kaspdb, true,
+	                          CTL_FILTER_BACKUP_KASPDB, CTL_FILTER_BACKUP_NOKASPDB) &&
+	    eval_opposite_filters(args, &filter_catalog, true,
+	                          CTL_FILTER_BACKUP_CATALOG, CTL_FILTER_BACKUP_NOCATALOG))) {
+		return KNOT_EXPARAM;
+	}
+
+	bool forced = ctl_has_flag(args->data[KNOT_CTL_IDX_FLAGS], CTL_FLAG_FORCE);
+
 	zone_backup_ctx_t *ctx;
 
 	// The present timer db size is not up-to-date, use the maximum one.
-	conf_val_t timer_db_size = conf_db_param(conf(), C_TIMER_DB_MAX_SIZE,
-	                                         C_MAX_TIMER_DB_SIZE);
+	conf_val_t timer_db_size = conf_db_param(conf(), C_TIMER_DB_MAX_SIZE);
 
-	int ret = zone_backup_init(restore_mode,
+	int ret = zone_backup_init(restore_mode, forced,
 	                           args->data[KNOT_CTL_IDX_DATA],
 	                           knot_lmdb_copy_size(&args->server->kaspdb),
 	                           conf_int(&timer_db_size),
@@ -431,82 +480,129 @@ static int init_backup(ctl_args_t *args, bool restore_mode)
 	}
 
 	assert(ctx != NULL);
-	ctx->backup_journal = MATCH_AND_FILTER(args, CTL_FILTER_PURGE_JOURNAL);
-	ctx->backup_zonefile = !MATCH_AND_FILTER(args, CTL_FILTER_PURGE_ZONEFILE);
-	args->custom_ctx = ctx;
+	ctx->backup_zonefile = filter_zonefile;
+	ctx->backup_journal = filter_journal;
+	ctx->backup_timers = filter_timers;
+	ctx->backup_kaspdb = filter_kaspdb;
+	ctx->backup_catalog = filter_catalog;
 
-	if (args->data[KNOT_CTL_IDX_ZONE] == NULL) {
-		ctx->backup_global = true;
-		ret = global_backup(ctx, &args->server->catalog, NULL);
-		if (ret != KNOT_EOK) {
-			zone_backup_deinit(ctx);
-		}
+	zone_backups_add(&args->server->backup_ctxs, ctx);
+
+	return ret;
+}
+
+static zone_backup_ctx_t *latest_backup_ctx(ctl_args_t *args)
+{
+	// no need to mutex in this case
+	return (zone_backup_ctx_t *)TAIL(args->server->backup_ctxs.ctxs);
+}
+
+static int deinit_backup(ctl_args_t *args)
+{
+	return zone_backup_deinit(latest_backup_ctx(args));
+}
+
+static int zone_backup_cmd(zone_t *zone, ctl_args_t *args)
+{
+	zone_backup_ctx_t *ctx = latest_backup_ctx(args);
+	if (!ctx->restore_mode && ctx->failed) {
+		// No need to proceed with already faulty backup.
+		return KNOT_EOK;
+	}
+
+	if (zone->backup_ctx != NULL) {
+		log_zone_warning(zone->name, "backup or restore already in progress, skipping zone");
+		ctx->failed = true;
+		return KNOT_EPROGRESS;
+	}
+
+	zone->backup_ctx = ctx;
+	pthread_mutex_lock(&ctx->readers_mutex);
+	ctx->readers++;
+	pthread_mutex_unlock(&ctx->readers_mutex);
+	ctx->zone_count++;
+
+	int ret = schedule_trigger(zone, args, ZONE_EVENT_BACKUP, true);
+
+	if (ret == KNOT_EOK && !ctx->backup_global && (ctx->restore_mode || !ctx->failed)) {
+		ret = global_backup(ctx, zone->catalog, zone->name);
 	}
 
 	return ret;
 }
 
-static void deinit_backup(ctl_args_t *args)
+static int zones_apply_backup(ctl_args_t *args, bool restore_mode)
 {
-	zone_backup_ctx_t *ctx = args->custom_ctx;
-	zone_backup_deinit(ctx);
+	int ret_deinit;
+	int ret = init_backup(args, restore_mode);
+
+	if (ret != KNOT_EOK) {
+		char *msg = sprintf_alloc("%s init failed (%s)",
+		                          restore_mode ? "restore" : "backup",
+		                          knot_strerror(ret));
+
+		if (args->data[KNOT_CTL_IDX_ZONE] == NULL) {
+			log_ctl_error("%s", msg);
+		} else {
+			log_ctl_zone_str_error(args->data[KNOT_CTL_IDX_ZONE],
+			                       "%s", msg);
+		}
+		free (msg);
+
+		/* Warning: zone name in the control command params discarded here. */
+		args->data[KNOT_CTL_IDX_ZONE] = NULL;
+		send_error(args, knot_strerror(ret));
+		return KNOT_CTL_EZONE;
+	}
+
+	/* Global catalog zones backup. */
+	if (args->data[KNOT_CTL_IDX_ZONE] == NULL) {
+		zone_backup_ctx_t *ctx = latest_backup_ctx(args);
+		ctx->backup_global = true;
+		ret = global_backup(ctx, &args->server->catalog, NULL);
+		if (ret != KNOT_EOK) {
+			log_ctl_error("control, error (%s)", knot_strerror(ret));
+			send_error(args, knot_strerror(ret));
+			ret = KNOT_EOK;
+			goto done;
+		}
+	}
+
+	ret = zones_apply(args, zone_backup_cmd);
+
+done:
+	ret_deinit = deinit_backup(args);
+	return ret != KNOT_EOK ? ret : ret_deinit;
 }
 
-static int zone_backup_cmd(zone_t *zone, ctl_args_t *args)
+static int zone_sign(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	zone_backup_ctx_t *ctx = args->custom_ctx;
-	if (zone->backup_ctx != NULL) {
-		log_zone_warning(zone->name, "backup already in progress");
-		args->failed = true;
-		return KNOT_EPROGRESS;
-	}
-	zone->backup_ctx = ctx;
-	pthread_mutex_lock(&ctx->readers_mutex);
-	ctx->readers++;
-	pthread_mutex_unlock(&ctx->readers_mutex);
-
-	schedule_trigger(zone, args, ZONE_EVENT_BACKUP, true);
-
-	if (ctx->backup_global) {
-		return KNOT_EOK;
-	} else {
-		return global_backup(ctx, zone->catalog, zone->name);
-	}
-}
-
-static int zone_sign(zone_t *zone, ctl_args_t *args)
-{
-	UNUSED(args);
-
 	conf_val_t val = conf_zone_get(conf(), C_DNSSEC_SIGNING, zone->name);
 	if (!conf_bool(&val)) {
+		args->suppress = true;
 		return KNOT_ENOTSUP;
 	}
 
 	zone_set_flag(zone, ZONE_FORCE_RESIGN);
-	schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, true);
 }
 
-static int zone_keys_load(zone_t *zone, ctl_args_t *args)
+static int zone_keys_load(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	conf_val_t val = conf_zone_get(conf(), C_DNSSEC_SIGNING, zone->name);
 	if (!conf_bool(&val)) {
+		args->suppress = true;
 		return KNOT_ENOTSUP;
 	}
 
-	schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, true);
 }
 
 static int zone_key_roll(zone_t *zone, ctl_args_t *args)
 {
 	conf_val_t val = conf_zone_get(conf(), C_DNSSEC_SIGNING, zone->name);
 	if (!conf_bool(&val)) {
+		args->suppress = true;
 		return KNOT_ENOTSUP;
 	}
 
@@ -519,15 +615,11 @@ static int zone_key_roll(zone_t *zone, ctl_args_t *args)
 		return KNOT_EINVAL;
 	}
 
-	schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, true);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, true);
 }
 
-static int zone_ksk_sbm_confirm(zone_t *zone, ctl_args_t *args)
+static int zone_ksk_sbm_confirm(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	kdnssec_ctx_t ctx = { 0 };
 
 	int ret = kdnssec_ctx_init(conf(), &ctx, zone->name, zone->kaspdb, NULL);
@@ -541,34 +633,24 @@ static int zone_ksk_sbm_confirm(zone_t *zone, ctl_args_t *args)
 	conf_val_t val = conf_zone_get(conf(), C_DNSSEC_SIGNING, zone->name);
 	if (ret == KNOT_EOK && conf_bool(&val)) {
 		// NOT zone_events_schedule_user(), intentionally!
-		schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, false);
+		ret = schedule_trigger(zone, args, ZONE_EVENT_DNSSEC, false);
 	}
 
 	return ret;
 }
 
-static int zone_freeze(zone_t *zone, ctl_args_t *args)
+static int zone_freeze(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
-	schedule_trigger(zone, args, ZONE_EVENT_UFREEZE, false);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_UFREEZE, false);
 }
 
-static int zone_thaw(zone_t *zone, ctl_args_t *args)
+static int zone_thaw(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
-	schedule_trigger(zone, args, ZONE_EVENT_UTHAW, false);
-
-	return KNOT_EOK;
+	return schedule_trigger(zone, args, ZONE_EVENT_UTHAW, false);
 }
 
-static int zone_txn_begin(zone_t *zone, ctl_args_t *args)
+static int zone_txn_begin(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	if (zone->control_update != NULL) {
 		return KNOT_TXN_EEXISTS;
 	}
@@ -583,20 +665,23 @@ static int zone_txn_begin(zone_t *zone, ctl_args_t *args)
 	if (ret != KNOT_EOK) {
 		free(zone->control_update);
 		zone->control_update = NULL;
-		return ret;
 	}
 
-	return KNOT_EOK;
+	return ret;
 }
 
-static int zone_txn_commit(zone_t *zone, ctl_args_t *args)
+static int zone_txn_commit(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	if (zone->control_update == NULL) {
+		args->suppress = true;
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
+	int ret = zone_update_semcheck(zone->control_update);
+	if (ret != KNOT_EOK) {
+		return ret; // Recoverable error.
+	}
+
 	// NOOP if empty changeset/contents.
 	if (((zone->control_update->flags & UPDATE_INCREMENTAL) &&
 	     changeset_empty(&zone->control_update->change)) ||
@@ -613,8 +698,8 @@ static int zone_txn_commit(zone_t *zone, ctl_args_t *args)
 		zone_sign_reschedule_t resch = { 0 };
 		bool full = (zone->control_update->flags & UPDATE_FULL);
 		zone_sign_roll_flags_t rflags = KEY_ROLL_ALLOW_ALL;
-		int ret = (full ? knot_dnssec_zone_sign(zone->control_update, 0, rflags, 0, &resch) :
-		                  knot_dnssec_sign_update(zone->control_update, &resch));
+		ret = (full ? knot_dnssec_zone_sign(zone->control_update, conf(), 0, rflags, 0, &resch) :
+		              knot_dnssec_sign_update(zone->control_update, conf(), &resch));
 		if (ret != KNOT_EOK) {
 			zone_control_clear(zone);
 			return ret;
@@ -622,7 +707,7 @@ static int zone_txn_commit(zone_t *zone, ctl_args_t *args)
 		event_dnssec_reschedule(conf(), zone, &resch, false);
 	}
 
-	int ret = zone_update_commit(conf(), zone->control_update);
+	ret = zone_update_commit(conf(), zone->control_update);
 	if (ret != KNOT_EOK) {
 		zone_control_clear(zone);
 		return ret;
@@ -636,11 +721,10 @@ static int zone_txn_commit(zone_t *zone, ctl_args_t *args)
 	return KNOT_EOK;
 }
 
-static int zone_txn_abort(zone_t *zone, ctl_args_t *args)
+static int zone_txn_abort(zone_t *zone, _unused_ ctl_args_t *args)
 {
-	UNUSED(args);
-
 	if (zone->control_update == NULL) {
+		args->suppress = true;
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
@@ -818,12 +902,14 @@ static int zone_read(zone_t *zone, ctl_args_t *args)
 			ret = zone_contents_nsec3_apply(zone->contents, send_node, ctx);
 		}
 	}
+
 	return ret;
 }
 
 static int zone_flag_txn_get(zone_t *zone, ctl_args_t *args, const char *flag)
 {
 	if (zone->control_update == NULL) {
+		args->suppress = true;
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
@@ -926,17 +1012,13 @@ static int send_changeset(changeset_t *ch, send_ctx_t *ctx)
 	}
 
 	// Second send 'to' changeset part.
-	ret = send_changeset_part(ch, ctx, false);
-	if (ret != KNOT_EOK) {
-		return ret;
-	}
-
-	return KNOT_EOK;
+	return send_changeset_part(ch, ctx, false);
 }
 
 static int zone_txn_diff(zone_t *zone, ctl_args_t *args)
 {
 	if (zone->control_update == NULL) {
+		args->suppress = true;
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
@@ -1045,6 +1127,7 @@ parser_failed:
 static int zone_txn_set(zone_t *zone, ctl_args_t *args)
 {
 	if (zone->control_update == NULL) {
+		args->suppress = true;
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
@@ -1068,6 +1151,7 @@ static int zone_txn_set(zone_t *zone, ctl_args_t *args)
 static int zone_txn_unset(zone_t *zone, ctl_args_t *args)
 {
 	if (zone->control_update == NULL) {
+		args->suppress = true;
 		return KNOT_TXN_ENOTEXISTS;
 	}
 
@@ -1217,40 +1301,51 @@ static int orphans_purge(ctl_args_t *args)
 
 static int zone_purge(zone_t *zone, ctl_args_t *args)
 {
+	int ret = KNOT_EOK;
+
 	// Abort possible editing transaction.
 	if (MATCH_OR_FILTER(args, CTL_FILTER_PURGE_EXPIRE)) {
-		(void)zone_txn_abort(zone, args);
+		ret = zone_txn_abort(zone, args);
+		RETURN_IF_FAILED(KNOT_TXN_ENOTEXISTS);
 	}
 
 	// Purge the zone timers.
 	if (MATCH_OR_FILTER(args, CTL_FILTER_PURGE_TIMERS)) {
 		memset(&zone->timers, 0, sizeof(zone->timers));
-		(void)zone_timers_sweep(&args->server->timerdb,
+		ret = zone_timers_sweep(&args->server->timerdb,
 		                        zone_names_distinct, zone->name);
+		RETURN_IF_FAILED(KNOT_ENOENT);
 	}
 
 	// Expire the zone.
 	if (MATCH_OR_FILTER(args, CTL_FILTER_PURGE_EXPIRE)) {
-		schedule_trigger(zone, args, ZONE_EVENT_EXPIRE, true);
+		// KNOT_EOK is the only return value from event_expire().
+		(void)schedule_trigger(zone, args, ZONE_EVENT_EXPIRE, true);
 	}
 
 	// Purge the zone file.
 	if (MATCH_OR_FILTER(args, CTL_FILTER_PURGE_ZONEFILE)) {
 		char *zonefile = conf_zonefile(conf(), zone->name);
-		(void)unlink(zonefile);
+		if (unlink(zonefile) == -1) {
+			ret = knot_map_errno();
+		}
 		free(zonefile);
+		RETURN_IF_FAILED(KNOT_ENOENT);
 	}
 
 	// Purge the zone journal.
 	if (MATCH_OR_FILTER(args, CTL_FILTER_PURGE_JOURNAL)) {
-		(void)journal_scrape_with_md(zone_journal(zone), true);
+		ret = journal_scrape_with_md(zone_journal(zone), true);
+		RETURN_IF_FAILED(KNOT_ENOENT);
 	}
 
 	// Purge KASP DB.
 	if (MATCH_OR_FILTER(args, CTL_FILTER_PURGE_KASPDB)) {
-		if (knot_lmdb_open(zone->kaspdb) == KNOT_EOK) {
-			(void)kasp_db_delete_all(zone->kaspdb, zone->name);
+		ret = knot_lmdb_open(zone->kaspdb);
+		if (ret == KNOT_EOK) {
+			ret = kasp_db_delete_all(zone->kaspdb, zone->name);
 		}
+		RETURN_IF_FAILED(KNOT_ENOENT);
 	}
 
 	return KNOT_EOK;
@@ -1407,7 +1502,6 @@ static int zone_stats(zone_t *zone, ctl_args_t *args)
 
 static int ctl_zone(ctl_args_t *args, ctl_cmd_t cmd)
 {
-	int ret;
 	switch (cmd) {
 	case CTL_ZONE_STATUS:
 		return zones_apply(args, zone_status);
@@ -1422,23 +1516,9 @@ static int ctl_zone(ctl_args_t *args, ctl_cmd_t cmd)
 	case CTL_ZONE_FLUSH:
 		return zones_apply(args, zone_flush);
 	case CTL_ZONE_BACKUP:
-		ret = init_backup(args, false);
-		if (ret == KNOT_EOK) {
-			ret = zones_apply(args, zone_backup_cmd);
-			deinit_backup(args);
-		} else {
-			send_error(args, knot_strerror(ret));
-		}
-		return ret;
+		return zones_apply_backup(args, false);
 	case CTL_ZONE_RESTORE:
-		ret = init_backup(args, true);
-		if (ret == KNOT_EOK) {
-			ret = zones_apply(args, zone_backup_cmd);
-			deinit_backup(args);
-		} else {
-			send_error(args, knot_strerror(ret));
-		}
-		return ret;
+		return zones_apply_backup(args, true);
 	case CTL_ZONE_SIGN:
 		return zones_apply(args, zone_sign);
 	case CTL_ZONE_KEYS_LOAD:
@@ -1496,7 +1576,7 @@ static int server_status(ctl_args_t *args)
 		ret = snprintf(buff, sizeof(buff), "Version: %s", PACKAGE_VERSION);
 	} else if (strcasecmp(type, "workers") == 0) {
 		int running_bkg_wrk, wrk_queue;
-		worker_pool_status(args->server->workers, &running_bkg_wrk, &wrk_queue);
+		worker_pool_status(args->server->workers, false, &running_bkg_wrk, &wrk_queue);
 		ret = snprintf(buff, sizeof(buff), "UDP workers: %zu, TCP workers: %zu, "
 		               "XDP workers: %zu, background workers: %zu (running: %d, pending: %d)",
 		               conf()->cache.srv_udp_threads, conf()->cache.srv_tcp_threads,
