@@ -1,4 +1,4 @@
-/*  Copyright (C) 2019 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2021 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -98,7 +98,8 @@ static int send_notify(conf_t *conf, zone_t *zone, const knot_rrset_t *soa,
 
 	const struct sockaddr_storage *dst = &slave->addr;
 	const struct sockaddr_storage *src = &slave->via;
-	knot_request_t *req = knot_request_make(NULL, dst, src, pkt, &slave->key, 0);
+	knot_request_flag_t flags = conf->cache.srv_tcp_fastopen ? KNOT_REQUEST_TFO : 0;
+	knot_request_t *req = knot_request_make(NULL, dst, src, pkt, &slave->key, flags);
 	if (!req) {
 		knot_request_free(req, NULL);
 		knot_requestor_clear(&requestor);
@@ -110,6 +111,7 @@ static int send_notify(conf_t *conf, zone_t *zone, const knot_rrset_t *soa,
 	if (ret == KNOT_EOK && knot_pkt_ext_rcode(req->resp) == 0) {
 		NOTIFY_OUT_LOG(LOG_INFO, zone->name, dst,
 		               "serial %u", knot_soa_serial(soa->rrs.rdata));
+		zone->timers.last_notified_serial = (knot_soa_serial(soa->rrs.rdata) | LAST_NOTIFIED_SERIAL_VALID);
 	} else if (knot_pkt_ext_rcode(req->resp) == 0) {
 		NOTIFY_OUT_LOG(LOG_WARNING, zone->name, dst,
 		               "failed (%s)", knot_strerror(ret));
@@ -129,6 +131,8 @@ int event_notify(conf_t *conf, zone_t *zone)
 {
 	assert(zone);
 
+	bool failed = false;
+
 	if (zone_contents_is_empty(zone->contents)) {
 		return KNOT_EOK;
 	}
@@ -143,16 +147,22 @@ int event_notify(conf_t *conf, zone_t *zone)
 		conf_val_t addr = conf_id_get(conf, C_RMT, C_ADDR, &notify);
 		size_t addr_count = conf_val_count(&addr);
 
+		int ret = KNOT_EOK;
+
 		for (int i = 0; i < addr_count; i++) {
 			conf_remote_t slave = conf_remote(conf, &notify, i);
-			int ret = send_notify(conf, zone, &soa, &slave, timeout);
+			ret = send_notify(conf, zone, &soa, &slave, timeout);
 			if (ret == KNOT_EOK) {
 				break;
 			}
 		}
 
+		if (ret != KNOT_EOK) {
+			failed = true;
+		}
+
 		conf_val_next(&notify);
 	}
 
-	return KNOT_EOK;
+	return failed ? KNOT_ERROR : KNOT_EOK;
 }
