@@ -207,6 +207,7 @@ int net_init(const srv_info_t     *local,
 	}
 
 	// Store network parameters.
+	net->sockfd = -1;
 	net->iptype = iptype;
 	net->socktype = socktype;
 	net->wait = wait;
@@ -658,6 +659,8 @@ void net_clean(net_t *net)
 
 	free(net->local_str);
 	free(net->remote_str);
+	net->local_str = NULL;
+	net->remote_str = NULL;
 
 	if (net->local_info != NULL) {
 		if (net->local == NULL) {
@@ -665,10 +668,12 @@ void net_clean(net_t *net)
 		} else {
 			freeaddrinfo(net->local_info);
 		}
+		net->local_info = NULL;
 	}
 
 	if (net->remote_info != NULL) {
 		freeaddrinfo(net->remote_info);
+		net->remote_info = NULL;
 	}
 
 #ifdef LIBNGHTTP2
