@@ -1,4 +1,4 @@
-/*  Copyright (C) 2020 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2021 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -636,7 +636,7 @@ static int cmd_zone_ctl(cmd_args_t *args)
 	return ctl_receive(args);
 }
 
-#define MAX_FILTERS 7
+#define MAX_FILTERS 12
 
 typedef struct {
 	const char *name;
@@ -649,9 +649,17 @@ const filter_desc_t zone_flush_filters[MAX_FILTERS] = {
 };
 
 const filter_desc_t zone_backup_filters[MAX_FILTERS] = {
-	{ "+backupdir",   CTL_FILTER_FLUSH_OUTDIR,   true },
-	{ "+journal",     CTL_FILTER_PURGE_JOURNAL,  false },
-	{ "+nozonefile",  CTL_FILTER_PURGE_ZONEFILE, false },
+	{ "+backupdir",   CTL_FILTER_BACKUP_OUTDIR,      true },
+	{ "+zonefile",    CTL_FILTER_BACKUP_ZONEFILE,   false },
+	{ "+nozonefile",  CTL_FILTER_BACKUP_NOZONEFILE, false },
+	{ "+journal",     CTL_FILTER_BACKUP_JOURNAL,    false },
+	{ "+nojournal",   CTL_FILTER_BACKUP_NOJOURNAL,  false },
+	{ "+timers",      CTL_FILTER_BACKUP_TIMERS,     false },
+	{ "+notimers",    CTL_FILTER_BACKUP_NOTIMERS,   false },
+	{ "+kaspdb",      CTL_FILTER_BACKUP_KASPDB,     false },
+	{ "+nokaspdb",    CTL_FILTER_BACKUP_NOKASPDB,   false },
+	{ "+catalog",     CTL_FILTER_BACKUP_CATALOG,    false },
+	{ "+nocatalog",   CTL_FILTER_BACKUP_NOCATALOG,  false },
 };
 
 const filter_desc_t zone_status_filters[MAX_FILTERS] = {
@@ -664,9 +672,9 @@ const filter_desc_t zone_status_filters[MAX_FILTERS] = {
 
 const filter_desc_t zone_purge_filters[MAX_FILTERS] = {
 	{ "+expire",   CTL_FILTER_PURGE_EXPIRE },
-	{ "+timers",   CTL_FILTER_PURGE_TIMERS },
 	{ "+zonefile", CTL_FILTER_PURGE_ZONEFILE },
 	{ "+journal",  CTL_FILTER_PURGE_JOURNAL },
+	{ "+timers",   CTL_FILTER_PURGE_TIMERS },
 	{ "+kaspdb",   CTL_FILTER_PURGE_KASPDB },
 	{ "+orphan",   CTL_FILTER_PURGE_ORPHAN },
 };
